{-# LANGUAGE FlexibleContexts, FlexibleInstances, MultiParamTypeClasses, FunctionalDependencies, UndecidableInstances #-}
module Fp11 where
--import Control.Monad.Instances
import Control.Applicative
--import Control.Monad.Identity
--import Data.Monoid
--import Data.Maybe
--import Control.Monad.Writer
--import Control.Monad.Reader
--import Control.Monad.State
import Data.Char
import Control.Monad.Cont
import Control.Monad.State
import Control.Monad.Identity

-- depricated начиная c ghc 7.10 
import Control.Monad.Error
-- будет Control.Monad.Except без малоосмысленного Error

-- *Fp11> Nothing <|> (Just 3) <|> (Just 5) <|> Nothing
-- Just 3


{-
> import Text.Parsec
> let parser = string "ABC" <|> string "DEF"
> parseTest parser "ABC123"
"ABC"
> parseTest parser "DEF123"
"DEF"
> parseTest parser "GHI123"
parse error at (line 1, column 1):
unexpected "G"
expecting "ABC" or "DEF"
-}



-- *Fp11> match (sym 'a' <|> sym 'b') "a"
-- Just 'a'
-- *Fp11> match (many $ psym isAlpha) "abc"
-- Just "abc"
-- *Fp11> match (many $ psym isAlpha) "123"
-- Nothing
-- *Fp11> match (many $ psym isAlpha) ""
-- Just ""
-- *Fp11> match (some $ psym isAlpha) ""
-- Nothing

-- 
-- *Fp11> Nothing `mplus` (Just 3) `mplus` (Just 5) `mplus` Nothing
-- Just 3

pythags = do
  z <- [1..]
  x <- [1..z]
  y <- [x..z]
  guard $ x^2 + y^2 == z^2
  return (x, y, z)

-- *Fp11> take 5 pythags
-- [(3,4,5),(6,8,10),(5,12,13),(9,12,15),(8,15,17)]

--------------------------------------------------

data Vector = Vector Int Int 
               deriving (Eq, Show)
data Matrix = Matrix Vector Vector 
               deriving (Eq, Show)
               
class Mult a b c | a b -> c where
  (***) :: a -> b -> c
 
instance Mult Matrix Matrix Matrix where
  Matrix (Vector a11 a12) (Vector a21 a22) *** Matrix (Vector b11 b12) (Vector b21 b22) =
    Matrix (Vector c11 c12) (Vector c21 c22) where
      c11 = a11 * b11 + a12 * b21
      c12 = a11 * b12 + a12 * b22
      c21 = a21 * b11 + a22 * b21
      c22 = a21 * b12 + a22 * b22
 
-- *Fp11> let a = Matrix (Vector 1 2) (Vector 3 4)
-- *Fp11> let i = Matrix (Vector 1 0) (Vector 0 1)
-- *Fp11> a *** i
--  Matrix (Vector 1 2) (Vector 3 4)

--------------------------------------------------
-- MonadError
data DivByError = ErrZero | Other String
                   deriving (Eq, Read, Show)

instance Error DivByError where
  strMsg s = Other s
  noMsg = Other "Unknown DivByError"
 

(/?) :: Double -> Double -> Either DivByError Double
x /? 0 = throwError ErrZero
x /? y = return $ x / y

example0 :: Double -> Double -> Either DivByError String
example0 x y = action  `catchError` handler where 
  action = do 
    q <- x /? y
    return $ show q
  handler = \err -> return $ show err

{-
*Fp11> example0 5 2
Right "2.5"
*Fp11> example0 5 0
Right "ErrZero"
-}

-- требуется расширение FlexibleContexts
(?/?) :: (MonadError DivByError m) 
            => Double -> Double -> m Double
x ?/? 0 = throwError ErrZero
x ?/? y = return $ x / y

example1 :: Double -> Double -> Either DivByError String
example1 x y = action  `catchError` handler where 
  action = do 
    q <- x ?/? y
    return $ show q
  handler = \err -> return $ show err


-- в Control.Monad.Except для обеспечения такого поведения
-- тип, используемый как носитель ошибки, должен быть моноидом (вроде так :)
example2 :: Double -> Double -> Either DivByError String
example2 x y = action `catchError` handler where 
  action = do 
    q <- x ?/? y
    guard $ y>=0
    return $ show q
  handler = \err -> return $ show err

{-
*Fp11> example2 7 (-5)
Right "Other \"Unknown DivByError\""
-}

--------------------------------------------------------

stInteger :: State Integer Integer       
stInteger = do modify (+1)
               a <- get
               return a
         
stString :: State String String       
stString = do modify (++"1")
              b <- get
              return b

-- *Fp11> evalState stInteger 0
-- 1
-- *Fp11> evalState stString "0"
-- "01"

stComb :: StateT Integer (StateT String Identity) (Integer, String)      
stComb = do modify (+1)
            lift $ modify (++"1")
            a <- get
            b <- lift get
            return (a,b)

-- *Fp11> runIdentity $ evalStateT (evalStateT stComb 0) "0"
-- (1,"01")


newtype MaybeT m a = MaybeT { runMaybeT :: m (Maybe a) }

instance (Monad m) => Monad (MaybeT m) where
  fail _  = MaybeT $ return Nothing
  return  = lift . return
  x >>= f = MaybeT $ do
    v <- runMaybeT x
    case v of
      Nothing -> return Nothing
      Just y  -> runMaybeT (f y)

instance MonadTrans MaybeT where
  lift = MaybeT . liftM Just
  
mbSt :: MaybeT (StateT Integer Identity) Integer      
mbSt = do 
  lift $ modify (+1)
  a <- lift get
  True <- return $ a >= 3
  --guard (a >= 3)
  return a
{-  
*Fp11> runIdentity $ evalStateT (runMaybeT mbSt) 0
Nothing
*Fp11> runIdentity $ evalStateT (runMaybeT mbSt) 2
Just 3
-}

-- чтобы заработал guard нужно сделать MaybeT m представителем MonadPlus
instance (Monad m) =>  MonadPlus (MaybeT m) where
  mzero       = MaybeT $ return Nothing
  x `mplus` y = MaybeT $ do
    v <- runMaybeT x
    case v of
      Nothing -> runMaybeT y
      Just _  -> return v

mbSt' :: MaybeT (State Integer) Integer
mbSt' = do lift $ modify (+1)
           a <- lift get
           guard $ a >= 3 -- !!
           return a
{-
*Fp11> runIdentity $ evalStateT (runMaybeT mbSt') 0
Nothing
*Fp11> runIdentity $ evalStateT (runMaybeT mbSt') 2
Just 3
-}

-- Для любой пары монад можно избавиться от подъёма 
-- стандартных операций вложенной монады

{-
class Monad m => MonadState s m | m -> s where
    get :: m s
    get = state (\s -> (s, s))
    
    put :: s -> m ()
    put s = state (\_ -> ((), s))
    
    state :: (s -> (a, s)) -> m a
    state f = do
      s <- get
      let ~(a, s') = f s
      put s'
      return a

modify :: MonadState s m => (s -> s) -> m ()
modify f = state (\s -> ((), f s))

gets :: MonadState s m => (s -> a) -> m a
gets f = do
    s <- get
    return (f s)

-}

instance (MonadState s m) => MonadState s (MaybeT m) where
  get = lift get
  put = lift . put


mbSt'' :: MaybeT (State Integer) Integer
mbSt'' = do 
  modify (+1) -- без lift
  a <- get -- без lift
  guard $ a >= 3
  return a
{-
*Fp11> runIdentity $ evalStateT (runMaybeT mbSt'') 0
Nothing
*Fp11> runIdentity $ evalStateT (runMaybeT mbSt'') 2
Just 3
-}